#ifndef _XAIFBOOSTERBASICBLOCKPREACCUMULATIONTAPEADJOINT_BASICBLOCKALG_INCLUDE_
#define _XAIFBOOSTERBASICBLOCKPREACCUMULATIONTAPEADJOINT_BASICBLOCKALG_INCLUDE_
// ========== begin copyright notice ==============
// This file is part of 
// ---------------
// xaifBooster
// ---------------
// Distributed under the BSD license as follows:
// Copyright (c) 2005, The University of Chicago
// All rights reserved.
//
// Redistribution and use in source and binary forms, 
// with or without modification, are permitted provided that the following conditions are met:
//
//    - Redistributions of source code must retain the above copyright notice, 
//      this list of conditions and the following disclaimer.
//    - Redistributions in binary form must reproduce the above copyright notice, 
//      this list of conditions and the following disclaimer in the documentation 
//      and/or other materials provided with the distribution.
//    - Neither the name of The University of Chicago nor the names of its contributors 
//      may be used to endorse or promote products derived from this software without 
//      specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
// EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
// OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
// SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
// INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 
// General Information:
// xaifBooster is intended for the transformation of 
// numerical programs represented as xml files according 
// to the XAIF schema. It is part of the OpenAD framework. 
// The main application is automatic 
// differentiation, i.e. the generation of code for 
// the computation of derivatives. 
// The following people are the principal authors of the 
// current version: 
// 	Uwe Naumann
//	Jean Utke
// Additional contributors are: 
//	Andrew Lyons
//	Peter Fine
//
// For more details about xaifBooster and its use in OpenAD please visit:
//   http://www.mcs.anl.gov/openad
//
// This work is partially supported by:
// 	NSF-ITR grant OCE-0205590
// ========== end copyright notice ==============

#include "xaifBooster/system/inc/Assignment.hpp"
#include "xaifBooster/system/inc/BaseConstant.hpp"
#include "xaifBooster/system/inc/SubroutineCall.hpp"
#include "xaifBooster/system/inc/Symbol.hpp"
#include "xaifBooster/system/inc/PlainBasicBlock.hpp"
#include "xaifBooster/system/inc/ForLoopReversalType.hpp"

#include "xaifBooster/algorithms/DerivativePropagator/inc/DerivativePropagatorEntry.hpp"

#include "xaifBooster/algorithms/TypeChange/inc/TemporariesHelper.hpp"

#include "xaifBooster/algorithms/BasicBlockPreaccumulationTape/inc/BasicBlockAlg.hpp"

#include "xaifBooster/algorithms/InlinableXMLRepresentation/inc/InlinableSubroutineCall.hpp"

using namespace xaifBooster;

namespace xaifBoosterBasicBlockPreaccumulationTapeAdjoint {  

  /** 
   * class to implement a reinterpretation of 
   * the DerivativePropagator instances as propagators in the adjoint sweep 
   */
  class BasicBlockAlg : public xaifBoosterBasicBlockPreaccumulationTape::BasicBlockAlg {
  public:
    
    BasicBlockAlg(BasicBlock& theContaining);

    virtual ~BasicBlockAlg();

    virtual void printXMLHierarchy(std::ostream& os) const;

    virtual std::string debug() const ;

    virtual void traverseToChildren(const GenericAction::GenericAction_E anAction_c);
    
    xaifBoosterInlinableXMLRepresentation::InlinableSubroutineCall& addInlinableSubroutineCall(const std::string& aSubroutineName,
											       const ForLoopReversalType::ForLoopReversalType_E& aReversalType);

    SubroutineCall& addSubroutineCall(const Symbol& aSubroutineNameSymbol,
				      const Scope& aSubroutineNameScope,
				      ActiveUseType::ActiveUseType_E anActiveUse,
				      const ForLoopReversalType::ForLoopReversalType_E& aReversalType,
                                      unsigned short formalArgCount);

    virtual void algorithm_action_5();

    const PlainBasicBlock::BasicBlockElementList& getBasicBlockElementList(const ForLoopReversalType::ForLoopReversalType_E& aReversalType) const; 
    
  private:

    /// no def
    BasicBlockAlg();

    /// no def
    BasicBlockAlg(const BasicBlockAlg&);

    /// no def
    BasicBlockAlg& operator=(const BasicBlockAlg&);

    typedef std::pair<const Variable*, const Variable*> VariablePPair;
    typedef std::list<VariablePPair> VariablePPairList;

    /// This list maps propagation address variables with the corresponding variables that have been popped
    VariablePPairList myAddressVariableCorList;

    /// This list maps propagation factor variables with the corresponding variables that have been popped
    VariablePPairList myFactorVariableCorList;

    /** 
     * the elements that make up the adjoint code
     * which assumes an anonymous reversal (no knowledge about the original 
     * loop variables)
     */
    PlainBasicBlock::BasicBlockElementList myBasicBlockElementListAnonymousReversal;

    /** 
     * the elements that make up the adjoint code
     * under an explicit reversal (loop constructs are reversed explicitly
     * and we assume all index expressions can be recalculated explcitly 
     * at reversal time from explicitly reversed loops)
     */
    PlainBasicBlock::BasicBlockElementList myBasicBlockElementListExplicitReversal;
    
    /** 
     * popping array index values if we pushed any
     * based on the reversal type (see also ControlFlowReversal)
     */
    void reinterpretArrayAccess(const Variable& theOriginalVariable,
				Variable& theActualVariable,
				const ForLoopReversalType::ForLoopReversalType_E& aReversalType);

    /** 
     * based on the reversal type (see also ControlFlowReversal)
     */
    void reinterpretDerivativePropagatorEntry(const xaifBoosterDerivativePropagator::DerivativePropagatorEntry& aDerivativePropagatorEntry);

    PlainBasicBlock::BasicBlockElementList& getBasicBlockElementList(const ForLoopReversalType::ForLoopReversalType_E& aReversalType); 

    /** 
     * add the call to the proper BasicBlockElementList based on aReversalType
     */
    void addZeroDeriv(Variable& theTarget,
		      const ForLoopReversalType::ForLoopReversalType_E& aReversalType);

    /** 
     * add the call to the proper BasicBlockElementList based on aReversalType
     */
    void addUnitFactor(Variable& theSource,
		       Variable& theTarget,
		       const ForLoopReversalType::ForLoopReversalType_E& aReversalType);

    /** 
     * add the call to the proper BasicBlockElementList based on aReversalType
     */
    void addNegativeUnitFactor(Variable& theSource,
                               Variable& theTarget,
                               const ForLoopReversalType::ForLoopReversalType_E& aReversalType);

    /** 
     * add the call to the proper BasicBlockElementList based on aReversalType
     */
    const Variable& addFactorPop(const Symbol& aTemporarySymbol,
				 const ForLoopReversalType::ForLoopReversalType_E& aReversalType);

    /** 
     * add the call to the proper BasicBlockElementList based on \p aReversalType
     * Unlike addFactorPop, this method generates its own symbol for the variable to be popped into.
     */
    const Variable& addAddressPop(const ForLoopReversalType::ForLoopReversalType_E& aReversalType);

    /** 
     * add the call to allocate the temporary symbol
     */
    void addAllocation(const Symbol& aTemporarySymbol,
		       const Scope& theScope,
		       xaifBoosterTypeChange::TemporariesHelper& aHelper,
		       const ForLoopReversalType::ForLoopReversalType_E& aReversalType);

    /** 
     * add a Saxpy call with variable factor \p theFactor to the proper BasicBlockElementList based on aReversalType
     */
    void addSaxpy(const Variable& theSource,
		  const Variable& theTarget,
		  const Variable& theFactor,
		  const ForLoopReversalType::ForLoopReversalType_E& aReversalType);

    /** 
     * add a Saxpy call with constant factor \p theConstantFactor to the proper BasicBlockElementList based on aReversalType
     */
    void addSaxpy(const Variable& theSource,
                  const Variable& theTarget,
                  const Constant& theFactor,
                  const ForLoopReversalType::ForLoopReversalType_E& aReversalType);

  }; // end class xaifBoosterBasicBlockPreaccumulationTapeAdjoint::BasicBlockAlg
 
} // end namespace xaifBoosterBasicBlockPreaccumulationTapeAdjoint
                                                                     
#endif
